import textwrap
from dataclasses import asdict, dataclass
from typing import Any, Dict, List, Optional, Union, cast

from typing_extensions import Literal

from airplane.api.entities import BuiltInRun
from airplane.files import File, upload as file_upload
from airplane.runtime import __execute_internal


@dataclass
class MessageOption:
    """The options for sending a Slack message.

    This includes the same options as the Slack API's `chat.postMessage` method.
    """

    attachments: Optional[List[Dict]] = None
    text: Optional[str] = None
    blocks: Optional[List[Dict]] = None
    mrkdwn: Optional[bool] = None
    parse: Optional[Literal["full", "none"]] = None
    reply_broadcast: Optional[bool] = None
    thread_ts: Optional[str] = None
    unfurl_links: Optional[bool] = None
    unfurl_media: Optional[bool] = None


def message(
    channel_name: str,
    message: Union[str, MessageOption],  # pylint: disable=redefined-outer-name
    dedent: bool = True,
) -> BuiltInRun[None]:
    """Sends a message to a Slack channel.

    Args:
        channel_name: The Slack channel to send a message to. This can be a channel name (e.g.
            `#general`), a channel ID (e.g. `C1234567890`), or a user ID (e.g. `U1234567890`).
        message: The message to send. This can be a string, or a Slack message option object. The
            Slack message option object includes the same options as the Slack API's
            `chat.postMessage` method.
        dedent: Whether or not to omit leading whitespace from `message`.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the message builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """
    param_values: Dict[str, Any]
    if isinstance(message, str):
        if dedent:
            message = textwrap.dedent(message)
        param_values = {
            "channelName": channel_name,
            "message": message,
        }
    else:
        param_values = {
            "channelName": channel_name,
            "message": "",
            "messageOption": asdict(message),
        }
    return cast(
        BuiltInRun[None],
        __execute_internal(
            "airplane:slack_message",
            param_values,
            {"slack": "res00000000zteamslack"},
        ),
    )


def upload(
    channel_name: str,
    payload: Union[bytes, str, File],
    filename: str,
    message: Optional[str] = None,  # pylint: disable=redefined-outer-name
) -> BuiltInRun[None]:
    """Uploads a file to a Slack channel.

    Args:
        channel_name: The Slack channel to send a message to. This can be a channel name (e.g.
            `#general`), a channel ID (e.g. `C1234567890`), or a user ID (e.g. `U1234567890`).
        payload: Payload to upload to Slack. Can be a string, bytes, or an Airplane file.
        filename: Name of the upload. This is used to determine the file type of the upload.
        message: Optional message to send with the upload.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the message builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """
    if isinstance(payload, File):
        file_url = payload.url
    else:
        file_url = file_upload(payload, filename).url
    param_values = {
        "channelName": channel_name,
        "fileURL": file_url,
        "fileName": filename,
    }
    if message is not None:
        param_values["message"] = message
    return cast(
        BuiltInRun[None],
        __execute_internal(
            "airplane:slack_upload",
            param_values,
            {"slack": "res00000000zteamslack"},
        ),
    )



from typing import Optional, Union

import filetype
import requests

from airplane.api.client import api_client_from_env
from airplane.types import File


def upload(payload: Union[bytes, str], file_name: Optional[str] = None) -> File:
    """Upload a payload to Airplane and create an Airplane File object.

    Args:
        payload: Content of the file to upload.
        file_name: Name of the file to create. If not provided, the filename will
            be generated by trying to infer the file type from the payload. If
            the file type cannot be inferred, an error will be raised.

    Returns:
        An Airplane File object.

    Raises:
        HTTPError: If the file could not be uploaded.
        ValueError: If the file type could not be inferred and no filename was
            provided.
    """
    if isinstance(payload, str):
        payload = payload.encode("utf-8")
        if file_name is None:
            file_name = "file.txt"
    else:
        if file_name is None:
            kind = filetype.guess(payload)
            if kind is None:
                raise ValueError(
                    "Unable to determine file type, please specify a filename"
                )
            file_name = f"file.{kind.extension}"
    upload_response = api_client_from_env().create_upload(file_name, len(payload))

    try:
        response = requests.put(
            upload_response["writeOnlyURL"],
            data=payload,
            headers={"X-Goog-Content-Length-Range": f"0,{len(payload)}"},
            timeout=600,
        )
        response.raise_for_status()
    except requests.HTTPError as err:
        raise requests.HTTPError(
            f"Failed to upload file: {err.response.text}", response=err.response
        ) from err

    return File(
        id=upload_response["upload"]["id"],
        url=upload_response["readOnlyURL"],
    )



from enum import Enum
from typing import Any, Dict, Optional, Union, cast

from typing_extensions import Literal, TypedDict

from airplane.api.entities import BuiltInRun
from airplane.builtins import __convert_resource_alias_to_id
from airplane.runtime import __execute_internal


class Method(Enum):
    """Valid HTTP methods for REST requests."""

    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"


class BodyType(Enum):
    """Valid HTTP body types for REST requests."""

    UNKNOWN = ""
    JSON = "json"
    RAW = "raw"
    FORM_DATA = "form-data"
    FORM_URL_ENCODED = "x-www-form-urlencoded"


class RequestOutput(TypedDict):
    """The output of the rest.request builtin."""

    status_code: int
    response: Union[str, Dict[str, Any]]


def request(
    rest_resource: str,
    method: Literal["GET", "POST", "PUT", "PATCH", "DELETE"],
    path: str,
    headers: Optional[Dict[str, Any]] = None,
    url_params: Optional[Dict[str, Any]] = None,
    body_type: BodyType = BodyType.UNKNOWN,
    body: Optional[Union[Dict[str, Any], str]] = None,
    form_data: Optional[Dict[str, Any]] = None,
    retry_failures: bool = False,
) -> BuiltInRun[RequestOutput]:
    """Runs the builtin request function against a REST Airplane resource.

    Args:
        rest_resource: The alias of the REST resource to use.
        method: The HTTP method of the request.
        path: The path of the request.
        headers: Optional headers to include in the request.
        url_params: Optional url params to include in the request.
        body_type: The type of the body if provided.
        body: The request body of type body_type to include in the request.
        form_data: The form data to include in the request.
        retry_failures: True to retry 500, 502, 503, and 504 error codes.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the request builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """
    if isinstance(method, Method):
        method_str = method.value
    else:
        method_str = method
    return cast(
        BuiltInRun[RequestOutput],
        __execute_internal(
            "airplane:rest_request",
            {
                "method": method_str,
                "path": path,
                "headers": headers,
                "urlParams": url_params,
                "bodyType": body_type.value,
                "body": body,
                "formData": form_data,
                "retryFailures": retry_failures,
            },
            {"rest": __convert_resource_alias_to_id(rest_resource)},
        ),
    )



import textwrap
from enum import Enum
from typing import Any, Dict, Optional, cast

from airplane.api.entities import BuiltInRun
from airplane.builtins import __convert_resource_alias_to_id
from airplane.runtime import __execute_internal


class TransactionMode(Enum):
    """Valid transaction modes for SQL Airplane resources."""

    AUTO = "auto"
    READ_ONLY = "readOnly"
    READ_WRITE = "readWrite"
    NONE = "none"


def query(
    sql_resource: str,
    query: str,  # pylint: disable=redefined-outer-name
    query_args: Optional[Dict[str, Any]] = None,
    transaction_mode: TransactionMode = TransactionMode.AUTO,
    dedent: bool = True,
) -> BuiltInRun[Dict[str, Any]]:
    """Runs the builtin query function against a SQL Airplane resource.

    Args:
        sql_resource: The alias of the SQL resource to execute the query against.
        query: The query to run on the SQL resource. Multiple queries can be separated by
            semicolons.
        query_args: Optional map of query arg names to values to insert into the query.
        transaction_mode: Optional transaction mode with which to run the query.
        dedent: Whether or not to omit leading whitespace from `query`.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

        The outputs of the run will be a map of query index to query result. For example,
        if you execute two queries::

            result = sql.query(
                "my_db",
                '''
                    SELECT email FROM users;
                    SELECT name FROM teams;
                ''',
            )

        `result.output` will be a dictionary with keys `Q1` and `Q2`.

        `result.output["Q1"]` will contain a dictionary of `{"email": ...}` from the
        `SELECT email FROM users;` query.

        `result.output["Q2"]` will contain a dictionary of `{"name": ...}` from the
        `SELECT name FROM teams;` query.

        In the case that your query is a DDL query without a `RETURNING` clause, the
        output will be a dictionary with key `rows_affected` containing the number of
        rows affected by the query.

    Raises:
        HTTPError: If the query builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """
    if dedent:
        query = textwrap.dedent(query)
    return cast(
        BuiltInRun[Dict[str, Any]],
        __execute_internal(
            "airplane:sql_query",
            {
                "query": query,
                "queryArgs": query_args,
                "transactionMode": transaction_mode.value,
            },
            {"db": __convert_resource_alias_to_id(sql_resource)},
        ),
    )



"""airplane - An SDK for writing Airplane tasks in Python"""

from airplane import auth, display, files, sleep
from airplane._version import __version__
from airplane.api.client import APIClient
from airplane.api.entities import PromptReviewers, Run, RunStatus
from airplane.builtins import ai, email, graphql, mongodb, rest, slack, sql
from airplane.config import (
    EnvVar,
    ExplicitPermissions,
    PermissionAssignees,
    Resource,
    Schedule,
    Webhook,
    task,
)
from airplane.exceptions import (
    InvalidEnvironmentException,
    PromptCancelledError,
    RunPendingException,
)
from airplane.output import append_output, set_output, write_named_output, write_output
from airplane.params import LabeledOption, ParamConfig
from airplane.runtime import execute, prompt
from airplane.runtime.standard import run  # Deprecated
from airplane.types import JSON, SQL, ConfigVar, File, LongText

__all__ = [
    "auth",
    "display",
    "files",
    "sleep",
    "__version__",
    "APIClient",
    "PromptReviewers",
    "Run",
    "RunStatus",
    "ai",
    "email",
    "graphql",
    "mongodb",
    "rest",
    "slack",
    "sql",
    "EnvVar",
    "ExplicitPermissions",
    "PermissionAssignees",
    "Resource",
    "Schedule",
    "Webhook",
    "task",
    "InvalidEnvironmentException",
    "PromptCancelledError",
    "RunPendingException",
    "append_output",
    "set_output",
    "write_named_output",
    "write_output",
    "LabeledOption",
    "ParamConfig",
    "execute",
    "prompt",
    "run",
    "JSON",
    "SQL",
    "ConfigVar",
    "File",
    "LongText",
]



from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, Generic, List, Optional, TypeVar

from airplane.types import JSONType


class RunStatus(Enum):
    """Valid statuses during a run's lifecycle."""

    NOT_STARTED = "NotStarted"
    QUEUED = "Queued"
    ACTIVE = "Active"
    SUCCEEDED = "Succeeded"
    FAILED = "Failed"
    CANCELLED = "Cancelled"

    def is_terminal(self) -> bool:
        """Returns whether the status is terminal.

        Returns:
            Whether the status is terminal.
        """

        return self in [self.SUCCEEDED, self.FAILED, self.CANCELLED]


JSONTypeT = TypeVar(
    "JSONTypeT",
    bound=JSONType,
)


@dataclass
class BuiltInRun(Generic[JSONTypeT]):
    """Representation of an Airplane built-in run.

    Attributes:
        id: The id of the run.
        param_values: The param values the run was provided.
        status: The current status of the run.
        output: The outputs (if any) of the run.
    """

    id: str
    param_values: Dict[str, Any]
    status: RunStatus
    output: JSONTypeT


@dataclass
class Run:
    """Representation of an Airplane run.

    Attributes:
        id: The id of the run.
        task_id: The task id associated with the run (None for builtin tasks).
        param_values: The param values the run was provided.
        status: The current status of the run.
        output: The outputs (if any) of the run.
    """

    id: str
    task_id: Optional[str]
    param_values: Dict[str, Any]
    status: RunStatus
    output: JSONType


@dataclass
class PromptReviewers:
    """Reviewers that are allowed to approve the prompt.

    Args:
        groups: List of groups allowed to approve the prompt. Groups are
            referenced via their slugs.
        users: List of users allowed to approve the prompt. Users are
            referenced via their emails.
        allow_self_approvals: Whether or not the run creator is allowed to approve
            their own prompt.
    """

    groups: Optional[List[str]] = None
    users: Optional[List[str]] = None
    allow_self_approvals: bool = True


@dataclass
class TaskReviewer:
    """Reviewers that are allowed to approve the task.

    Args:
        group_id: The ID of the group allowed to approve the task.o
        user_id: The ID of the user allowed to approve the task.
    """

    group_id: Optional[str] = None
    user_id: Optional[str] = None


@dataclass
class Sleep:
    """Airplane sleep object."""

    id: str
    run_id: str
    created_at: str
    until: str
    duration_ms: int
    skipped_at: Optional[str]
    skipped_by: Optional[str]


@dataclass
class User:
    """Airplane user."""

    id: str
    email: str
    name: str



import dataclasses
import datetime
import os
from enum import Enum
from typing import Any, Dict, List, Optional, Union, overload

import inflection
from typing_extensions import Literal

from airplane.api.entities import PromptReviewers, Run, User
from airplane.params import (
    SERIALIZED_DATE_FORMAT,
    SERIALIZED_DATETIME_FORMAT,
    SERIALIZED_DATETIME_MILLISECONDS_FORMAT,
    Constraints,
    ParamConfig,
    PromptParamTypes,
    SerializedParam,
    make_options,
    resolve_type,
    serialize_param,
    to_serialized_airplane_type,
)
from airplane.runtime.standard import (
    execute as standard_execute,
    get_prompt as standard_get_prompt,
    get_user as standard_get_user,
    prompt_background as standard_prompt_background,
    wait_for_prompt as standard_wait_for_prompt,
)
from airplane.types import ConfigVar, File

_AIRPLANE_RUNTIME_ENV_VAR = "AIRPLANE_RUNTIME"


class RuntimeKind(Enum):
    """Valid runtime kinds for Airplane runs."""

    DEV = "dev"
    STANDARD = ""
    WORKFLOW = "workflow"


def execute(slug: str, param_values: Optional[Dict[str, Any]] = None) -> Run:
    """Executes an Airplane task, waits for execution, and returns run metadata.

    Args:
        slug: The slug of the task to run.
        param_values: Optional map of parameter slugs to values.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the task cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
        NotImplementedError: For workflow runs.
    """
    return __execute_internal(slug, param_values)


def __execute_internal(
    slug: str,
    param_values: Optional[Dict[str, Any]] = None,
    resources: Optional[Dict[str, Any]] = None,
) -> Run:
    runtime_kind = os.environ.get(_AIRPLANE_RUNTIME_ENV_VAR, RuntimeKind.STANDARD.value)

    if runtime_kind == RuntimeKind.WORKFLOW.value:
        raise NotImplementedError("Workflow run not supported yet by python sdk")

    return standard_execute(slug, param_values, resources)


@dataclasses.dataclass
class Prompt:
    """Representation of an Airplane prompt."""

    prompt_id: str
    parameters: List[SerializedParam]

    def wait(self) -> Dict[str, Any]:
        """Waits on an operator for prompt input.

        Raises:
            NotImplementedError: For workflow runs.

        Example:
            Prompt user to input a refund amount and reason::

                prompt = airplane.prompt(
                    {"amount": float, "reason": str},
                    background=True,
                )
                values = prompt.wait()
                # Access values as `values["amount"]` and `values["reason"]`
        """
        runtime_kind = os.environ.get(
            _AIRPLANE_RUNTIME_ENV_VAR, RuntimeKind.STANDARD.value
        )

        if runtime_kind == RuntimeKind.WORKFLOW.value:
            raise NotImplementedError("Workflow run not supported yet by python sdk")

        prompt_info = standard_wait_for_prompt(self.prompt_id)
        prompt_values = prompt_info["values"]
        for param in self.parameters:
            # If the user didn't provide a value for the parameter slug
            if param.slug not in prompt_values:
                # Fill in None values for optional parameters that aren't provided
                if param.constraints.optional:
                    prompt_values[param.slug] = None
                # We should never get here...
                continue

            if param.multi:
                prompt_values[param.slug] = [
                    _convert_prompt_param(param, value)
                    for value in prompt_values[param.slug]
                ]
            else:
                prompt_values[param.slug] = _convert_prompt_param(
                    param,
                    prompt_values[param.slug],
                )

        return prompt_values

    def submitter(self) -> Optional[User]:
        """Returns the user who submitted the prompt, if there is one."""
        runtime_kind = os.environ.get(
            _AIRPLANE_RUNTIME_ENV_VAR, RuntimeKind.STANDARD.value
        )

        if runtime_kind == RuntimeKind.WORKFLOW.value:
            raise NotImplementedError("Workflow run not supported yet by python sdk")

        prompt_info = standard_get_prompt(self.prompt_id)
        if prompt_info.get("submittedBy") is None:
            return None

        user_info = standard_get_user(prompt_info["submittedBy"])
        return User(
            id=user_info["userID"],
            email=user_info["email"],
            name=user_info["name"],
        )


def _convert_prompt_param(param: SerializedParam, value: Any) -> Any:
    if param.type == "date":
        return datetime.datetime.strptime(value, SERIALIZED_DATE_FORMAT).date()
    if param.type == "datetime":
        return datetime.datetime.strptime(
            value,
            SERIALIZED_DATETIME_MILLISECONDS_FORMAT
            if "." in value
            else SERIALIZED_DATETIME_FORMAT,
        )
    if param.type == "upload":
        return File(
            id=value["id"],
            url=value["url"],
        )
    if param.type == "configvar":
        return ConfigVar(
            name=value["name"],
            value=value["value"],
        )

    return value


@overload
def prompt(
    params: Optional[Dict[str, PromptParamTypes]] = None,
    *,
    reviewers: Optional[PromptReviewers] = None,
    confirm_text: Optional[str] = None,
    cancel_text: Optional[str] = None,
    description: Optional[str] = None,
    background: Literal[False] = False,
    notify: bool = True,
) -> Dict[str, Any]:
    ...


@overload
def prompt(
    params: Optional[Dict[str, PromptParamTypes]] = None,
    *,
    reviewers: Optional[PromptReviewers] = None,
    confirm_text: Optional[str] = None,
    cancel_text: Optional[str] = None,
    description: Optional[str] = None,
    background: Literal[True],
    notify: bool = True,
) -> Prompt:
    ...


def prompt(
    params: Optional[Dict[str, PromptParamTypes]] = None,
    *,
    reviewers: Optional[PromptReviewers] = None,
    confirm_text: Optional[str] = None,
    cancel_text: Optional[str] = None,
    description: Optional[str] = None,
    background: bool = False,
    notify: bool = True,
) -> Union[Dict[str, Any], Prompt]:
    """Prompts an operator for input.

    Each prompt contains a parameter form, similar to what you see when executing a task or runbook.
    When a prompt is created, the run pauses and waits for a response before continuing.
    By default, anyone who can access the associated run can also respond to the prompt,
    but this can be restricted by specifying reviewers in the prompt's options. As a convenience,
    you can use `confirm` to create prompts with no parameters.

    Args:
        params: Defines the parameter form as a mapping of slug to parameter.
            The values will be returned as an object that maps the parameter slug to the
            corresponding value. To learn more about parameters, see
            the [Parameter documentation](https://docs.airplane.dev/platform/parameters).
        reviewers: Reviewers that are allowed to approve the prompt. If no reviewers are provided,
            anyone that can see the active run is allowed to approved.
        confirm_text: Text of the confirmation button on the prompt dialog.
        cancel_text: Text of the cancellation button on the prompt dialog.
        description: Prompt description to display. Supports markdown.
        background: If true, the prompt will be created in the background and a
            `Prompt` object will be returned. To wait for the prompt to be submitted,
            call `prompt.wait()`.
        notify: If true, the prompt will send a notification to the reviewers.

    Raises:
        NotImplementedError: For workflow runs.

    Example:
        Prompt user to input a refund amount and reason::

            values = airplane.prompt({
                "amount": int,
                "reason": str,
            }, confirm_text="Accept refund")
            # Access values as `values["amount"]` and `values["reason"]`

         Prompt with a list of options to choose from. As a convenience, you can use
         `prompt_select` to create a prompt with a list of objects::

             values = airplane.prompt({
                "username": Annotated[
                    str,
                    prompt.options(["colin", "eric"])
                ],
             })
             # Access values as `values["username"]`
    """

    runtime_kind = os.environ.get(_AIRPLANE_RUNTIME_ENV_VAR, RuntimeKind.STANDARD.value)

    if runtime_kind == RuntimeKind.WORKFLOW.value:
        raise NotImplementedError("Workflow run not supported yet by python sdk")

    serialized_params = []
    for slug, param in (params or {}).items():
        param_info = resolve_type(slug, param)
        param_config = param_info.param_config
        if param_config is None:
            param_config = ParamConfig()

        default = param_config.default
        if default is not None:
            default = serialize_param(default)

        serialized_type, component = to_serialized_airplane_type(
            slug, param_info.resolved_type
        )
        slug = param_config.slug or slug
        serialized_params.append(
            SerializedParam(
                slug=slug,
                name=param_config.name or inflection.humanize(slug),
                type=serialized_type,
                component=component,
                desc=param_config.description,
                default=default,
                multi=param_info.is_multi,
                constraints=Constraints(
                    optional=param_info.is_optional,
                    options=make_options(param_config),
                    regex=param_config.regex,
                ),
            )
        )

    prompt_id = standard_prompt_background(
        serialized_params,
        reviewers=reviewers,
        confirm_text=confirm_text,
        cancel_text=cancel_text,
        description=description,
        notify=notify,
    )
    background_prompt = Prompt(prompt_id, serialized_params)

    if background:
        return background_prompt

    return background_prompt.wait()



import dataclasses
import datetime
import functools
import inspect
from collections import Counter
from typing import Any, Callable, Dict, List, Optional, Union

import inflection
import typing_extensions
from docstring_parser import parse
from typing_extensions import Literal, ParamSpec

from airplane._version import __version__
from airplane.api.entities import Run
from airplane.exceptions import (
    InvalidAnnotationException,
    InvalidTaskConfigurationException,
    UnsupportedDefaultTypeException,
)
from airplane.params import (
    SERIALIZED_DATE_FORMAT,
    SERIALIZED_DATETIME_FORMAT_WITH_TIMEZONE,
    SERIALIZED_DATETIME_MILLISECONDS_FORMAT,
    ParamConfig,
    ParamDefOptions,
    ParamDefTypes,
    ParamType,
    ParamTypes,
    make_options,
    resolve_type,
    serialize_param,
    to_airplane_type,
)
from airplane.runtime import execute
from airplane.types import ConfigVar, File, RuntimeType
from airplane.utils import make_slug

# Restrict task execution so it can only be called from other tasks or views.
TaskCaller = Literal["task", "view"]

DefaultRunPermission = Literal["task-viewers", "task-participants"]


@dataclasses.dataclass(frozen=True)
class Resource:
    """Airplane resource attachment.

    Resources in Airplane allow users to configure connections to external systems
    like databases and APIs and use them in tasks and runbooks.
    https://docs.airplane.dev/resources/overview

    Attributes:
        slug:
            Resource identifier.
        alias:
            Alias to reference the resource. Defaults to resource slug.
    """

    slug: str
    alias: Optional[str] = None


@dataclasses.dataclass(frozen=True)
class EnvVar:
    """Airplane environment variable.

    Environment variables allow users to configure constant values or reference
    config variables. They can be accessed via `os.getenv("MY_ENV_VAR_NAME")`.

    Attributes:
        name:
            Environment variable name.
        value:
            Set a constant value for the environment variable.
        config_var_name:
            Name of the config variable to use. Configs variables allow users
            to set team-wide values / secrets.
    """

    name: str
    value: Optional[str] = None
    config_var_name: Optional[str] = None

    def __post_init__(self) -> None:
        if "=" in self.name:
            raise InvalidTaskConfigurationException(
                f'Environment variable name {self.name} cannot contain "="'
            )
        if (self.config_var_name and self.value) or (
            not self.config_var_name and not self.value
        ):
            raise InvalidTaskConfigurationException(
                "Exactly one of `config_var_name` or `value` should be set"
            )


@dataclasses.dataclass(frozen=True)
class Schedule:
    """Airplane schedule definition.

    Schedules allow users to automatically run tasks on a recurring schedule.
    https://docs.airplane.dev/schedules/schedules

    Attributes:
        slug:
            Human-friendly identifier used for the schedule. Schedule slugs must be unique
            within an individual task / workflow.
        cron:
            Schedule cron expression, e.g.  "0 * * * *"
        name:
            Schedule name. Defaults to the slug.
        description:
            Schedule description displayed on the Airplane app.
        param_values:
            Dictionary of parameter name to parameter value used for the scheduled execution.
            All required parameters must be defined.
    """

    slug: str
    cron: str
    name: Optional[str] = None
    description: Optional[str] = None
    param_values: Optional[Dict[str, Optional[ParamTypes]]] = None


@dataclasses.dataclass(frozen=True)
class Webhook:
    """Airplane webhook definition.

    Webhooks allow users to trigger tasks via HTTP requests.

    Attributes:
        slug:
            Human-friendly identifier used for the webhook. Webhook slugs must be unique
            within an individual task / workflow.
        require_airplane_token:
            Require an Airplane API key to be passed in the `X-Airplane-API-Key` header.
            Defaults to False.
    """

    slug: str
    require_airplane_token: bool = False


@dataclasses.dataclass
class PermissionAssignees:
    """Airplane permission assignees.

    Attributes:
        users:
            Users to assign the permission to. Users are referenced via their emails.
        groups:
            Groups to assign the permission to. Groups are referenced via their slugs.
    """

    users: Optional[List[str]] = None
    groups: Optional[List[str]] = None


@dataclasses.dataclass(frozen=True)
class ExplicitPermissions:
    """Airplane explicit permissions.

    Explicit task permissions in Airplane allow users to configure granular group-based
    or user-based permissions for the task:
    https://docs.airplane.dev/platform/permissions#task-and-runbook-permissions

    Attributes:
        viewers:
            Groups and users who can see task information, but can't request or execute tasks.
        requesters:
            Groups and users who have all the permission of viewers, and can also request tasks.
        executers:
            Groups and users who have all the permissions of requesters, and can also execute
            tasks and other's requests.
        admins:
            Groups and users who have full access to the task, and can change task configurations
            and permissions.
    """

    viewers: Optional[PermissionAssignees] = None
    requesters: Optional[PermissionAssignees] = None
    executers: Optional[PermissionAssignees] = None
    admins: Optional[PermissionAssignees] = None


P = ParamSpec("P")


def task(
    slug: Optional[str] = None,
    name: Optional[str] = None,
    description: Optional[str] = None,
    require_requests: bool = False,
    allow_self_approvals: bool = True,
    restrict_callers: Optional[List[TaskCaller]] = None,
    default_run_permissions: Optional[DefaultRunPermission] = None,
    timeout: int = 3600,
    concurrency_key: Optional[str] = None,
    concurrency_limit: Optional[int] = None,
    constraints: Optional[Dict[str, str]] = None,
    resources: Optional[List[Resource]] = None,
    schedules: Optional[List[Schedule]] = None,
    webhooks: Optional[List[Webhook]] = None,
    env_vars: Optional[List[EnvVar]] = None,
    permissions: Optional[Union[Literal["team_access"], ExplicitPermissions]] = None,
) -> Callable[[Callable[P, Any]], Callable[P, Run]]:
    """Decorator used to define an Airplane task.

    This decorator inspects the decorated function to create an Airplane task. The task's parameters
    are deduced from the function's arguments and type hints.

    Additional parameter metadata can be provided by attaching an airplane.ParamConfig to a
    typing.Annotated typehint (for versions of Python prior to 3.9, typing_extensions.Annotated
    can be used).

    The following map enumerates the supported Python types and equivalent Airplane
    parameter types (Airplane parameter details https://docs.airplane.dev/platform/parameters)::

        str: short text
        airplane.LongText: long text
        airplane.SQL: SQL
        bool: boolean
        int: integer
        float: number
        airplane.File: file
        datetime.date: date
        datetime.datetime: datetime
        airplane.ConfigVar: config variable

    Example:
        Sample task::

            @airplane.task()
            def add_two_numbers(first_number: int, second_number: int) -> int:
                return first_number + second_number

        Sample task with parameter annotation::

            @airplane.task()
            def capitalize_string(
                input: Annotated[str, airplane.ParamConfig(name="User string")]
            ) -> str:
                return input.capitalize()
    Args:
        slug:
            Human-friendly identifier used to reference this task. Must be unique
            across tasks and workflows. Defaults to function name.
        name:
            Task name displayed on the Airplane app. Defaults to funcion name in sentence case.
        description:
            Task description displayed on the Airplane app. If not provided, the description
            will be pulled from the docstring of the decorated function.
        require_requests:
            Whether or not this task requires a request to execute.
        allow_self_approvals:
            Whether or not this task allows self approvals.
        restrict_callers:
            Restrict task execution to specific callers. This disables direct execution
            and hides the task in the UI.
        default_run_permissions:
            Manage who has permissions for new runs of this task. If not provided,
            defaults to "task-viewers".
        timeout:
            How long a task can run (in seconds) for before it is automatically cancelled.
        concurrency_key:
            Restricts runs with the same concurrency key from executing at the same time.
        concurrency_limit:
            If concurrency key is set, only allows this task's runs to start if the number of
            other active runs with the same key is below this limit.
        constraints:
            Constraints for which agents are allowed to execute this task's runs, only
            applicable for users with self hosted agents.
        resources:
            Resources to attach to this task. Resources can be accessed through environment
            variables or built-ins. Resources accessed by this task must be explicitly attached
            in the task's definition.
        schedules:
            Schedules to attach to this task. Schedules allow users to automatically run
            task on a recurring schedule.
        webhooks:
            Webhooks to attach to this task. Webhooks allow users to trigger tasks
            from external systems via HTTP request.
        env_vars:
            Enviornment variables to attach to this task. Environment variables allow users
            to configure constant values or reference config variables.
        permissions:
            Permissions determine the users and groups that can access the task. Permissions can
            be set to "team_access" to allow full access to anyone on your team or to explicit
            permissions to configure granular group-based or user-based permissions for the task.
    """

    def decorator(func: Callable[P, Any]) -> Callable[P, Run]:
        """Assigns an __airplane attribute to a function to mark it as an Airplane object"""

        config = TaskDef.build(
            func=func,
            runtime="",
            slug=slug,
            name=name,
            description=description,
            require_requests=require_requests,
            allow_self_approvals=allow_self_approvals,
            restrict_callers=restrict_callers,
            timeout=timeout,
            default_run_permissions=default_run_permissions,
            concurrency_key=concurrency_key,
            concurrency_limit=concurrency_limit,
            constraints=constraints,
            resources=resources,
            schedules=schedules,
            webhooks=webhooks,
            env_vars=env_vars,
            permissions=permissions,
        )

        @functools.wraps(func)
        def wrapped(*args: P.args, **kwargs: P.kwargs) -> Run:
            kwargs.update(zip(func.__code__.co_varnames, args))  # type: ignore
            return execute(config.slug, kwargs)

        # pylint: disable=protected-access
        wrapped.__airplane = config  # type: ignore
        return wrapped

    return decorator


@dataclasses.dataclass(frozen=True)
class ParamDef:
    """Parameter definition"""

    # Represents the original function's argument name.
    # We need this in order to be able to map slug -> arg_name in case they differ.
    arg_name: str
    slug: str
    name: str
    type: ParamType
    description: Optional[str]
    default: Optional[ParamDefTypes]
    required: Optional[bool]
    multi: Optional[bool]
    options: Optional[ParamDefOptions]
    regex: Optional[str]


@dataclasses.dataclass(frozen=True)
class TaskDef:
    """Task definition"""

    func: Callable[..., Any]
    slug: str
    name: str
    runtime: RuntimeType
    entrypoint_func: str
    description: Optional[str]
    require_requests: Optional[bool]
    allow_self_approvals: Optional[bool]
    restrict_callers: Optional[List[TaskCaller]]
    timeout: Optional[int]
    concurrency_key: Optional[str]
    concurrency_limit: Optional[int]
    default_run_permissions: Optional[DefaultRunPermission]
    constraints: Optional[Dict[str, str]]
    resources: Optional[List[Resource]]
    schedules: Optional[List[Schedule]]
    parameters: Optional[List[ParamDef]]
    env_vars: Optional[List[EnvVar]]
    permissions: Optional[Union[Literal["team_access"], ExplicitPermissions]]
    sdk_version: str = dataclasses.field(default=__version__, init=False)
    webhooks: Optional[List[Webhook]]

    def run(self, params: Dict[str, Any]) -> Any:
        """Execute task function from param dictionary"""
        func_args: Dict[str, Any] = {}
        for param in self.parameters or []:
            # If the user didn't provide a value for the parameter slug
            if param.slug not in params:
                # Fill in None values for optional parameters that aren't provided
                if not param.required:
                    func_args[param.arg_name] = None
                # Otherwise, we fall back to the function default arguments.
                continue

            if param.multi:
                func_args[param.arg_name] = [
                    _convert_task_param(param, value) for value in params[param.slug]
                ]
            else:
                func_args[param.arg_name] = _convert_task_param(
                    param, params[param.slug]
                )

        return self.func(**func_args)

    @classmethod
    def build(
        cls,
        func: Callable[P, Any],
        runtime: RuntimeType,
        slug: Optional[str],
        name: Optional[str],
        description: Optional[str],
        require_requests: bool,
        allow_self_approvals: bool,
        restrict_callers: Optional[List[TaskCaller]],
        timeout: int,
        concurrency_key: Optional[str],
        concurrency_limit: Optional[int],
        default_run_permissions: Optional[DefaultRunPermission],
        constraints: Optional[Dict[str, str]],
        resources: Optional[List[Resource]],
        schedules: Optional[List[Schedule]],
        webhooks: Optional[List[Webhook]],
        env_vars: Optional[List[EnvVar]],
        permissions: Optional[Union[Literal["team_access"], ExplicitPermissions]],
    ) -> "TaskDef":
        """Construct a task definition from a function."""
        task_description = description
        if func.__doc__ is None:
            param_descriptions = {}
        else:
            docstring = parse(func.__doc__)
            param_descriptions = {
                param.arg_name: param.description for param in docstring.params
            }
            task_description = (
                description or docstring.long_description or docstring.short_description
            )
        type_hints = typing_extensions.get_type_hints(func, include_extras=True)
        sig = inspect.signature(func)
        parameters: List[ParamDef] = []
        for param in sig.parameters.values():
            type_hint = type_hints.get(param.name)
            if type_hint is None:
                raise InvalidAnnotationException(
                    prefix="Missing type annotation",
                    func_name=func.__name__,
                    param_name=param.name,
                )
            param_info = resolve_type(
                param.name,
                type_hint,
                func_name=func.__name__,
            )
            param_config = param_info.param_config
            if param_config is None:
                param_config = ParamConfig()

            if param.default is inspect.Signature.empty:
                default = None
            else:
                if isinstance(param.default, File):
                    raise UnsupportedDefaultTypeException(
                        "File defaults are not currently supported with inline code configuration."
                    )
                default = serialize_param(param.default)
            if param_config.default is not None and param_config.default != default:
                raise InvalidTaskConfigurationException(
                    f"Function {func.__name__} contains an invalid default value configuration "
                    f"for parameter {param.name}. Default values should be set via the Python "
                    "native default function parameter syntax, e.g. "
                    '`def my_task(my_param: str = "default")` instead of inside `ParamConfig`.'
                )

            default_slug = make_slug(param.name)
            parameters.append(
                ParamDef(
                    arg_name=param.name,
                    # Parameter slug is the parameter's name in snakecase
                    slug=param_config.slug or default_slug,
                    name=param_config.name or inflection.humanize(default_slug),
                    type=to_airplane_type(
                        param.name, param_info.resolved_type, func.__name__
                    ),
                    description=param_config.description
                    or param_descriptions.get(param.name),
                    required=not param_info.is_optional,
                    default=default,
                    multi=param_info.is_multi,
                    options=make_options(param_config),
                    regex=param_config.regex,
                )
            )

        def _check_duplicates(counter: Counter, duplicate_type: str) -> None:
            duplicates = [slug for slug, count in counter.items() if count > 1]
            if duplicates:
                raise InvalidTaskConfigurationException(
                    f"Function {func.__name__} has duplicate {duplicate_type} {duplicates}"
                )

        _check_duplicates(Counter([p.slug for p in parameters]), "parameter slugs")
        _check_duplicates(Counter([s.slug for s in schedules or []]), "schedule slugs")
        _check_duplicates(Counter([e.name for e in env_vars or []]), "env var names")

        # Convert schedule param values to the correct default types
        for schedule in schedules or []:
            if schedule.param_values is not None:
                for param_name, param_value in schedule.param_values.items():
                    if param_value is not None:
                        schedule.param_values[param_name] = serialize_param(param_value)

        # pylint: disable=protected-access
        default_slug = make_slug(func.__name__)
        return cls(
            func=func,
            runtime=runtime,
            slug=slug or default_slug,
            name=name or inflection.humanize(default_slug),
            description=task_description,
            require_requests=require_requests,
            allow_self_approvals=allow_self_approvals,
            restrict_callers=restrict_callers,
            timeout=timeout,
            concurrency_key=concurrency_key,
            concurrency_limit=concurrency_limit,
            default_run_permissions=default_run_permissions,
            constraints=constraints,
            schedules=schedules,
            resources=resources,
            parameters=parameters,
            env_vars=env_vars,
            permissions=permissions,
            entrypoint_func=func.__name__,
            webhooks=webhooks,
        )


def _convert_task_param(param: ParamDef, value: Any) -> Any:
    if param.type == "date":
        return datetime.datetime.strptime(value, SERIALIZED_DATE_FORMAT).date()
    if param.type == "datetime":
        return datetime.datetime.strptime(
            value,
            SERIALIZED_DATETIME_MILLISECONDS_FORMAT
            if "." in value
            else SERIALIZED_DATETIME_FORMAT_WITH_TIMEZONE,
        )
    if param.type == "upload":
        value = File(
            id=value["id"],
            url=value["url"],
        )
    if param.type == "configvar":
        return ConfigVar(
            name=value["name"],
            value=value["value"],
        )

    return value
